/* bb_reorg.c: cleanup the brett database.
    Copyright (C) 1996 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "config.h"
#include "libquark.h"
#include "brett_db.h"
#include "brett_int.h"

/*
BB_FUNCTION(Funktion,BOOLEAN,bb_reorg,
[[(void)]],[[
Diese Funktion reorganisiert die Brettdatenbank. Ihr Aufruf ist von Zeit
zu Zeit sinnvoll, um mittlerweile ungltige / berschriebene Eintrge
aus der xbrett.dat zu entfernen (in der neuere Daten einfach angehngt
werden).

Aus der brett.dat werden alle gelschten Bretter entfernt (und auch
alle aus anderen Grnden nicht gelesenen Bretter - das ist insbesondere
ntzlich, wenn sich die Datenbank ber doppelt vergebene Brettnummern
beschwert).

Diese Funktion darf nur aufgerufen werden wenn kein anderer Prozess
auf die Datenbank zugreift (sie versucht das durch locking zu 
gewhrleisten - aber das kann nicht alles abfangen).

Die Datenbank wird nicht berschrieben. Stattdessen liegen neue
Dateien in Verzeichnis, in dem die Datenbank liegt (brett.neu und
xbrett.neu). Der aufrufende Prozess ist dafr zustndig, die neuen
Dateien an die Stelle der alten zu schieben.

Liefert FALSE bei Mierfolg, TRUE bei Erfolg.
]]) */
BOOLEAN
bb_reorg(void)
{
	char fname[PATH_MAX];
	FILE *f;
	int i;
	int fd;
	char buf[1024];
	long groesse;
	size_t got;

	if (!bb_sys_dir)
		return FALSE;
	if (POS_INDEX(0) > sizeof(buf))
	{
		fprintf(stderr,
				"bb_reorg: buffer size mismatch. This should not happen\n");
		return FALSE;
	}

	groesse = lseek(bb_fh,0,SEEK_END);
	if (groesse==-1)
	{
		fprintf(stderr,"bb_reorg: cannot seek in brett.dat\n: %s",
				strerror(errno));
		return FALSE;
	}
#ifdef CAN_LOCK_FILES
	/* Ganze Datenbank locken. Das ist der letzte Sicherheitscheck, den
	 * wir machen knnen.
	 */
	if (0!=fd_readw_lock(bb_fh,0,SEEK_SET,groesse))
	{
		syslog(LOG_ERR,"Kann brettdatenbank nicht locken.");
		return FALSE;
	}
	/* brigens lassen wir den lock stehen, wenn die Funktion erfolgreich
	 * beendet wird.
	 */
#endif

	/* Zuerst mal die xbrett.dat */
	sprintf(fname,"%s/xbrett.neu",bb_sys_dir);
	f=fopen(fname,"w");
	if (!f)
	{
#ifdef CAN_LOCK_FILES
		fd_un_lock(bb_fh,0,SEEK_SET,groesse);
#endif
		fprintf(stderr,"bb_reorg: %s: %s\n",fname,strerror(errno));
		return FALSE;
	}
	FOR_EACH_BRETTNR(i)
	{
		BRETT *b=br_ind[i].brett;
		if (b && (b->userinfo || b->sysopinfo || b->lokalinfo || b->aliasname))
		{
			fprintf(f,"G%ld\n",(long) b->nr);
			if (b->userinfo)
				fprintf(f,"U%s\n",b->userinfo);
			if (b->sysopinfo)
				fprintf(f,"S%s\n",b->sysopinfo);
			if (b->aliasname)
				fprintf(f,"A%s\n",b->aliasname);
			if (b->lokalinfo)
			{
				char *p;
				char *s=b->lokalinfo;
				while ((p=strpbrk(s,"\r\n"))!=NULL)
				{
					fwrite(s,1,(p-s),f);
					while (isspace(*p))
						*p++=0;
					s=p;
				}
				if (s && *s)
					fwrite(s,1,strlen(s),f);
			}
		}
	}
	if (fclose(f)!=0)
	{
		fprintf(stderr,"bb_reorg: %s: %s\n",fname,strerror(errno));
		unlink(fname);
#ifdef CAN_LOCK_FILES
		fd_un_lock(bb_fh,0,SEEK_SET,groesse);
#endif
		return FALSE;
	}

	/*
	 * So, der einfache Teil wre erledigt. Jetzt zur brett.dat.
	 */
	sprintf(fname,"%s/brett.neu",bb_sys_dir);
	fd=open(fname,O_WRONLY|O_CREAT|O_TRUNC,0640);
	if (fd==-1)
	{
		fprintf(stderr,"bb_reorg: %s: %s\n",fname,strerror(errno));
#ifdef CAN_LOCK_FILES
		fd_un_lock(bb_fh,0,SEEK_SET,groesse);
#endif
		return FALSE;
	}

	/* Verwaltungsdaten kopieren */
	lseek(bb_fh,0,SEEK_SET);
	got=xread(bb_fh,buf,POS_INDEX(0));
	if (got != POS_INDEX(0)) {
		fprintf(stderr,"bb_reorg: cannot read from brett.dat: %s\n",
				got == -1 ? strerror(errno) : "Datei zu kurz");
		unlink(fname);
#ifdef CAN_LOCK_FILES
		fd_un_lock(bb_fh,0,SEEK_SET,groesse);
#endif
		close(fd);
		return FALSE;
	}
	if (xwrite(fd,buf,POS_INDEX(0))==-1)
	{
		fprintf(stderr,"bb_reorg: cannot write to %s: %s\n",
				fname, strerror(errno));
		unlink(fname);
#ifdef CAN_LOCK_FILES
		fd_un_lock(bb_fh,0,SEEK_SET,groesse);
#endif
		close(fd);
		return FALSE;
	}

	{
		BRETTNR bnr;
		int ind=0;
		int backup_fd;
		BOOLEAN ok=TRUE;
		backup_fd=bb_fh;
		bb_fh=fd;
		FOR_EACH_BRETTNR(bnr) {
			ULONG old_flags;
			if (!br_ind[bnr].brett)
				continue;
			br_ind[bnr].brett->index=ind++;
			old_flags= br_ind[bnr].brett->flags;

			/* bb_write_brett mu betrogen werden */
			br_ind[bnr].brett->flags|=B_LOCKED;
			if (!bb_write_brett(bnr,FALSE,FALSE))
			{
				ok=FALSE;
				break;
			}
			br_ind[bnr].brett->flags=old_flags;
		}
		bb_fh=backup_fd;
		if (!ok)
		{
			unlink(fname);
			close(fd);
#ifdef CAN_LOCK_FILES
			fd_un_lock(bb_fh,0,SEEK_SET,groesse);
#endif
			fprintf(stderr,"bb_reorg: cannot write to %s: %s\n",
					fname, strerror(errno));
			return FALSE;
		}
	}
	return TRUE;
}
