/* r_status.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>
#include <dirent.h>
#include <assert.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <errno.h>

#include "libquark.h"
#include "sdbm.h"
#include "prototyp.h"
#include "user.h"
#include "filelock.h"
#include "msg.h"
#include "routing.h"
#include "tompaths.h"
#include "box_conf.h"
#include "queues.h"
#include "crc32.h"


/*
 * Konzept:
 * - es gibt eine globale Datenbank, in der alle Statusmeldungen mit
 *   Absendern und Empfngern untergebracht sind. Die Daten sind in
 *   tom/system/mailstatus.dat untergebracht. 
 *   tom/system/mailstatus.{dir,pag} sind eine sdbm-Datenbank, die
 *   fr relativ schnellen Zugriff sorgt. (dbz wre hier *eindeutig*
 *   schneller, aber die sdbm ist ohnehin in die Quark eingelinkt).
 * - Jede Nachricht wird auch noch in eine userspezifische Datenbank
 *   eingetragen. "userspezifisch" heit hier, da auch fr andere
 *   Boxen und Gateways eine solche DB existiert. 
 *   Diese Datenbank ist ein ASCII-File und wird sequentiell gelesen 
 *   und geschrieben. Um Datenstze, z.B. den letzten Status, 
 *   berschreiben zu knnen, haben die letzten Zeilen eine feste
 *   Lnge.
 * - Im Prinzip ist die userspezifische Datenbank Overkill - es 
 *   wrde reichen, hier ein "spoolfile" pro User zu haben, in dem
 *   alle noch nicht zugestellten Stati stehen (und das nach dem 
 *   Tausch einfach gelscht wird). Das hat allerdings den kleinen
 *   Haken, da PSA ("alle Statusnachrichten senden") nicht funktioniert.
 *   Dafr wre es *deutlich* schneller.
 * - Apropos Performance: Das sequentielle berschreiben und Lesen
 *   der Statusmeldungen ist wahrscheinlich die ineffizienteste
 *   Routine in der ganzen Box. Mgliche Abhilfen:
 *   - pro User ein Index auf die Datei (sdbm, mal wieder). Mglichst
 *     erst ab einer bestimmten Gre. Nachteil: Putz wird etwas 
 *     langsamer. 
 *   - Statusmeldungen nur noch spoolen, PSA ignorieren (wie PSN
 *     behandeln), und bei Outfileverlust halt die Statusmeldungen
 *     verschwinden lassen.
 * - Putz: Der Putz ist im Prinzip ganz einfach - aus den userspezifischen
 *   DBs und der globalen .dat werden alle alten Eintrge entfernt
 *   (wobei die Vermeidung von Raceconditions recht interessant ist - 
 *   Dateien "in place" zu krzen setzt ftruncate voraus. Hoffen wir
 *   mal, da das berall funktioniert oder ersetzbar ist).
 *   Die sdbm-DB wird danach neu aufgebaut, es werden nicht die 
 *   einzelnen Eintrge berschrieben, wie es die Nachrichtendatenbank
 *   macht (der Grund, warum sie diesen performancefressenden Weg whlt
 *   ist, da ihre sdbm-DB Positionsangaben enthlt).
 */

#define SMODE_CLOSED 0
#define SMODE_SHARED 1
#define SMODE_EXCLUSIVE 2

extern USER user;

static int s_mode=SMODE_CLOSED;
static FILE *s_file;
static DBM *s_dbm;
static void r_st_close_global(void *);
static BOOLEAN r_st_open_global(BOOLEAN for_write);
static BOOLEAN r_st_write_one_status_to_user(FILE *f, const char *id, 
	const char *from, const char *to, const char *status, time_t modzeit, 
	time_t origzeit);
static BOOLEAN r_st_write_status_to_user(USERNR unr, const char *id, 
	const char *from, const char *to, const char *status, time_t modzeit, 
	time_t origzeit);
static void r_st_rebuild_db(FILE *);

static BOOLEAN
r_st_open_global(BOOLEAN for_write)
{
	char fname[PATH_MAX];
	char fname2[PATH_MAX];

	if (s_mode==SMODE_SHARED && for_write)
			r_st_close_global(NULL);
	if (s_mode!=SMODE_CLOSED)
		return TRUE;
		
	sprintf(fname,"%smailstatus.dat",box_conf.var_dir);
	s_file=fopen(fname,"r+b");
	if (!s_file)
		s_file=xfopen(fname,"w+b");

	if (for_write)
	{
		if (0!=fd_write_lock(fileno(s_file), 0,SEEK_SET,0))
		{
			/* die Datei konnte nicht gelockt werden */
			fclose(s_file);
			s_mode=SMODE_CLOSED;
			return FALSE;
		}
		s_mode=SMODE_EXCLUSIVE;
	}
	else
	{
		if (0!=fd_read_lock(fileno(s_file), 0,SEEK_SET,0))
		{
			/* die Datei konnte nicht gelockt werden */
			fclose(s_file);
			s_mode=SMODE_CLOSED;
			return FALSE;
		}
		s_mode=SMODE_SHARED;
	}

	sprintf(fname,"%smailstatus.dir",box_conf.var_dir);
	sprintf(fname2,"%smailstatus.pag",box_conf.var_dir);

	s_dbm = dbm_prep(fname,fname2,O_RDWR|O_CREAT,0640);
	XNULLCHECK(s_dbm); /* This must not happen */

	dqueue_add(QUEUE_EXIT,r_st_close_global,NULL);
	return TRUE;
}

static void
r_st_close_global(void *dummy TOM_ATTRIB_UNUSED)
{
	if (s_file)
		fclose(s_file);
	if (s_dbm)
		dbm_close(s_dbm);
	s_file=NULL;
	s_dbm=NULL;
	s_mode=SMODE_CLOSED;
}

static char *
normalize_id(const char *id)
{
	char *p;
	BOOLEAN at=FALSE;
	char *n=XMALLOC(strlen(id)+1);
	for (p=n;*id;p++,id++)
	{
		if (!at)
			*p=*id;
		else
			*p=tolower((unsigned char)*id);
		if (*id=='@')
			at=TRUE;
	}
	*p=0;
	return n;
}

/*
 * Diese Funktion schreibt eine neue Nachricht in die Statusdatenbank
 * (nicht einen Status, sondern die Daten ber die Nachricht). Sie 
 * wird aufgerufen, wenn eine neue Nachricht verschickt wurde.
 * Sie arbeitet mit Spooling, wenn nicht sofort in die DB geschrieben
 * werden kann, und arbeitet Spoolfiles ab, falls doch.
 */
BOOLEAN r_st_newmail(const char *id, const char *from, simplelist_t *to,
	time_t eingabezeit)
{
	char *kopie=NULL;
	size_t pos;
	datum key;
	datum set;
	simplelist_t *akt;
	DIR *d;
	struct dirent *de;
	char fname[PATH_MAX];
	char fname2[PATH_MAX];
	char buf[1024];
	ULONG crc32;

	/* Mit Writelock ffnen */
	if (!r_st_open_global(TRUE))
	{
		/* Kein Writelock erhltlich, dann spoolen */
		FILE *f;
		sprintf(fname,"%sstatus/nm%lu",box_conf.spool_dir,(ULONG) getpid());
		f=fopen(fname,"ab");
		if (!f)
		{
			sprintf(fname2,"%sstatus",box_conf.spool_dir);
			mkdir(fname2,0750);
			f=fopen(fname,"ab");
		}
		if (!f)
			return FALSE;

		fprintf(f,
			"#%s\n"
			"V%s\n"
			"E%lu\n",
				id,from ? from : "unbekannt" ,(unsigned long) eingabezeit);
		for (akt=to;akt != NULL;akt=akt->next)
			fprintf(f,"A%s\n",akt->data);
		/* wichtig: die e-Zeile gehrt ans Ende. Siehe unten */
		fprintf(f,"e%lu\n", (unsigned long) time(NULL));
		fclose(f);
		return TRUE;
	}

	if (id)
	{
		if (fseek(s_file,0,SEEK_END))
		{
			/* Was ist das? */
			r_st_close_global(NULL);
			return FALSE;
		}

		/* An die ID-Datei anhngen */
		pos=ftell(s_file);
		fprintf(s_file,
			"#%s\n"
			"V%s\n"
			"E%lu\n",
				id,from ? from : "unbekannt" ,(unsigned long) eingabezeit);
		for (akt=to;akt != NULL;akt=akt->next)
			fprintf(s_file,"A%s\n",akt->data);
		/* hier mu sie nicht ans Ende wie beim Spoolen, aber der Konsistenz
		 * wegen ...
		 */
		fprintf(s_file,"e%lu\n", (unsigned long) time(NULL));

		kopie = normalize_id (id);
		crc32=crc32ansi(kopie,strlen(kopie));

		key.dptr = (char *) &crc32;
		key.dsize = sizeof (crc32);

		set.dptr=(char *) &pos;
		set.dsize=sizeof(pos);

		if (dbm_store (s_dbm, key, set, DBM_INSERT)==-1)
		{
			/* Hm? */
			free(kopie);
			r_st_close_global(NULL);
			return FALSE;
		}
		free(kopie);
		kopie=NULL;
	}

	/* Da wir der Prozess sind, der einen Writelock hat, sprich die 
	 * Funktion des Servers hat, sollten wir uns nun mal darum kmmern, 
	 * gespoolte Statusnachrichten zu importieren.
	 */
	sprintf(fname,"%sstatus",box_conf.spool_dir);
	d=opendir(fname);
	if (!d)
		/* Hm? */
		return TRUE;
	while ((de=readdir(d))!=NULL)
	{
		FILE *f;
		pid_t pid;
		if (strncmp(de->d_name,"nm",2))
			continue;
		pid=strtol(de->d_name+2,NULL,10);
		if (!pid)
			continue;
		errno=0;
		if (kill(pid,0) == 0  || errno==EPERM)
			/* der Prozess existiert noch und schreibt vielleicht 
			 * gerade noch etwas in die Datei */
			continue;
		sprintf(fname,"%sstatus/%s",box_conf.spool_dir,de->d_name);

		f=xfopen(fname,"rb");
		unlink(fname); /* avoid unlikely, but possible race condition */

		pos=ftell(s_file);
		while (fgets(buf,sizeof(buf),f))
		{
			char *p=strpbrk(buf,"\r\n");
			if (p) *p=0;
			/* in die Datenbank schreiben */
			fprintf(s_file,"%s\n",buf);

			switch(*buf)
			{
				case '#':
					/* wenn nicht die erste Nachricht im Spoolfile */
					if (kopie)
					{
						/* in die sdbm eintragen */
						crc32=crc32ansi(kopie,strlen(kopie));

						key.dptr = (char *) &crc32;
						key.dsize = sizeof (crc32);

						set.dptr=(char *) &pos;
						set.dsize=sizeof(pos);

						if (dbm_store (s_dbm, key, set, DBM_INSERT)==-1)
						{
							/* Hm? */
							free(kopie);
							r_st_close_global(NULL);
							/* da gehen mglicherweise ein paar Status-
							 * nachrichen verloren (mindestens eine). Naja,
							 * was will man machen?  Hope this never
							 * happens ...
							 */
							closedir(d);
							fclose(f);
							return FALSE;
						}
						free(kopie);
					}
					kopie=normalize_id(buf+1);
					break;
				case 'e':
					/* Position der nchsten Nachricht merken. Die 'e'-Zeile
					 * ist immer die letzte einer Nachricht im Spoolfile
					 */
					pos=ftell(s_file);
					break;
			}
		}
		fclose(f);
		if (kopie)
		{
			crc32=crc32ansi(kopie,strlen(kopie));

			key.dptr = (char *) &crc32;
			key.dsize = sizeof (crc32);

			set.dptr=(char *) &pos;
			set.dsize=sizeof(pos);

			if (dbm_store (s_dbm, key, set, DBM_INSERT)==-1)
			{
				/* Hm? */
				free(kopie);
				r_st_close_global(NULL);
				/* da gehen mglicherweise ein paar Status-
				 * nachrichen verloren (mindestens eine). Naja,
				 * was will man machen?  Hope this never
				 * happens ...
				 */
				closedir(d);
				fclose(f);
				return FALSE;
			}
		}
	}
	closedir(d);
	return TRUE;
}

static BOOLEAN
r_st_set_one(const char *id, const char *from, const char *to, 
	const char *status, time_t zeit)
{
	size_t pos;
	char *id_kopie;
	char *to_kopie=NULL;
	char buf[1024];
	char *p;
	time_t originalzeit=-1;
	ULONG crc32;
	datum key;
	datum found;

	to_kopie=NULL;
	/* Wenn kein Empfnger angegeben ist, dann den Absender der
	 * Originalnachricht suchen */
	id_kopie = normalize_id (id);
	crc32=crc32ansi(id_kopie,strlen(id_kopie));
	free(id_kopie);

	key.dptr = (char *) &crc32;
	key.dsize = sizeof (crc32);

	found = dbm_fetch (s_dbm, key);

	if (!memcmp (&found, &nullitem, sizeof (nullitem)))
	{
		/* nicht gefunden. Den Status transportieren wir
		 * schonmal nicht mehr.
		 */
		return FALSE;
	}

    /* alignment! */
	memcpy (&pos, found.dptr, sizeof (pos));

	if (fseek (s_file, pos, SEEK_SET) != 0)
	{
	   /* Hm? */
		r_st_close_global(NULL);
		return FALSE;
	}

	if (!fgets (buf, sizeof (buf), s_file) || *buf != '#')
	{
	   /* Hm? Keine ID-Zeile? */
		r_st_close_global(NULL);
		return FALSE;
	}

	/* Die Daten der Nachricht aus der Datei lesen */	
	while (fgets (buf, sizeof (buf), s_file))
	{
		if (*buf == '#')
			break; /* das ist schon die nchste Nachricht */

		p = strpbrk (buf, "\r\n");
		if (p) *p = 0;

		switch (*buf)
		{
		case 'V':
			if (!to && !to_kopie)
				to_kopie=XSTRDUP(buf+1);
			break;
		case 'E':
			originalzeit = strtol (buf + 1, NULL, 10);
			break;
		}
	}
	if (originalzeit==(time_t) -1 || originalzeit==0)
		return FALSE;
	if (!to && !to_kopie)
		return FALSE;
	/*
	 * nun nach "to" oder "to_kopie" routen, und in die "mailstatus.dat"
	 * des Zielusers schreiben.
	 */
	 {
	 	r_adr_t *r=r_alloc();
	 	r->inputadr=XSTRDUP(to ? to : to_kopie);
	 	/* nunja, wir routen hier einen Status und nicht mehr */
	 	r->bezahlt=TRUE;
	 	r->sec_bezahlt=TRUE;
	 	r->level=ULV_MAX;
	 	r->sec_level=ULV_MAX;
	 	r->owner=user.nummer;
	 	if (!r_router(NULL,r))
	 	{
	 		syslog(LOG_NOTICE,"can't route status to %s: %s",to ? to : to_kopie,
	 			r->logstr ? r->logstr : "kein Fehlertext");
	 		r_free_chain(r);
	 		if (to_kopie)
	 			free(to_kopie);
	 		return FALSE;
	 	}
	 	r_st_write_status_to_user(r->user_gate_nr,
	 		id, from, to ? to : to_kopie, status, zeit, 
	 		originalzeit);
		r_free_chain(r);
	 	if (to_kopie)
	 		free(to_kopie);
	 }
	return TRUE;
}

/*
 * Diese Funktion wird aufgerufen, wenn ein neuer Status zu setzen
 * ist. "to" ist optional und fr Protokollerweiterungen im MausTausch
 * gedacht (man darf ja hoffen).
 */
BOOLEAN 
r_st_set_status(const char *id, const char *from, const char *to, 
	const char *status, time_t zeit)
{
	char fname[PATH_MAX];
	DIR *d;
	struct dirent *de;
	BOOLEAN ok;

	/* Mit Readlock ffnen */
	if (!r_st_open_global(FALSE))
	{
		/* Oh weia - wieder mal spoolen */
		FILE *f;
		sprintf(fname,"%sstatus/st%lu",box_conf.spool_dir,(ULONG) getpid());
		f=fopen(fname,"ab");
		if (!f)
			return FALSE;

		fprintf(f,
			"#%s\n"
			"V%s\n"
			"B%s\n"
			"E%lu\n",
				id,from ? from : "unbekannt" ,status,(unsigned long) zeit);
		if (to)
			fprintf(f,"A%s\n",to);
		fclose(f);
		return TRUE;
	}

	if (id)
		ok=r_st_set_one(id, from, to, status, zeit);
	else
		ok=TRUE;

	/* gespoolte Files abarbeiten */
	sprintf(fname,"%sstatus",box_conf.spool_dir);
	d=opendir(fname);
	if (!d)
		return ok;
	while ((de=readdir(d))!=NULL)
	{
		char *sid=NULL;
		char *sfrom=NULL;
		char *sto=NULL;
		char *sstatus=NULL;
		time_t szeit=-1;
		char buf[1024];
		FILE *f;
		pid_t pid;

		if (strncmp(de->d_name,"st",2))
			continue; /* falsches Paket */

		pid=strtol(de->d_name+2,NULL,10);
		if (!pid)
			continue;
		errno=0;
		if (kill(pid,0) == 0  || errno==EPERM)
			/* der Prozess existiert noch und schreibt vielleicht 
			 * gerade noch etwas in die Datei */
			continue;
		/* ok, this is our file */
		sprintf(fname,"%sstatus/%s",box_conf.spool_dir,de->d_name);

		f=fopen(fname,"rb");
		if (!f)
			continue;
		unlink(fname); /* Racecondition vermeiden */
		while (fgets(buf,sizeof(buf),f))
		{
			char *p=strpbrk(buf,"\r\n");
			if (p) *p=0;
			switch(*p)
			{
			case '#':
				if (sid)
				{
					if (szeit!=(time_t) -1 && sfrom && sstatus)
						ok&=r_st_set_one(sid, sfrom, sto, sstatus, szeit);
					free(sfrom);
					free(sto);
					free(sstatus);
					free(sid);
					szeit=-1;
				}
				sid=XSTRDUP(buf+1);
				break;
			case 'V':
				sfrom=XSTRDUP(buf+1);
				break;
			case 'A':
				sto=XSTRDUP(buf+1);
				break;
			case 'B':
				sstatus=XSTRDUP(buf+1);
				break;
			case 'E':
				szeit=strtol(buf+1,NULL,10);
				break;
			}
		}
		fclose(f);
		if (sid)
		{
			if (szeit!=(time_t) -1 && sfrom && sstatus)
				ok&=r_st_set_one(sid, sfrom, sto, sstatus, szeit);
			free(sfrom);
			free(sto);
			free(sstatus);
			free(sid);
			szeit=-1;
		}
	}
	closedir(d);
	return ok;
}


/*
 * Einen Status in die Datei des Users/Gates unr schreiben.
 * Datenformat:
 * #id
 * Vfrom
 * Ato
 * Bstatus   Bx\r\n              sprich 4 Zeichen.
 * Eorigzeit E10 Ziffern\r\n sprich 13 Zeichen.
 * Mmodzeit  13 Zeichen. (kommt beim Export in die B-Zeile)
 * Csetzzeit 13 Zeichen. ("jetzt")
 * Eventuell kann man spter eine Datenbank darber legen, aber 
 * fr den Anfang wird das reichen.
 */
static BOOLEAN
r_st_write_status_to_user(USERNR unr, const char *id, const char *from,
	const char *to, const char *status, time_t modzeit, time_t origzeit)
{
	char fname[PATH_MAX];
	FILE *f;
	DIR *d;
	struct dirent *de;
	BOOLEAN ok;

	sprintf(fname,"%s%lu.status",box_conf.tausch_dir,(ULONG) unr);
	f=fopen(fname,"r+b");
	if (!f)
		f=xfopen(fname,"w+b");

	if (0!=fd_write_lock(fileno(f), 0,SEEK_SET,0))
	{
		/* die Datei konnte nicht gelockt werden */
		fclose(f);
		/* spoolen im persnlichen Spoolverzeichnis */
		sprintf(fname,"%s%lu/st%lu",box_conf.spool_dir,(ULONG) unr, (ULONG) getpid());
		f=fopen(fname,"ab");
		if (!f)
			return FALSE;
		fprintf(f,
			"#%s\n"
			"V%s\n"
			"A%s\n"
			"B%s\n"
			"E%lu\n"
			"M%lu\n",
				id,from ? from : "unbekannt",to,status,(ULONG) origzeit, (ULONG) modzeit);
		fclose(f);
		return TRUE;
	}

	if (id)
		ok=r_st_write_one_status_to_user(f, id,from,to,status,modzeit,origzeit);
	else
		ok=TRUE;
	fclose(f);

	/* gespoolte Files abarbeiten */
	sprintf(fname,"%s%lu",box_conf.spool_dir,(ULONG) unr);
	d=opendir(fname);
	if (!d)
		return ok;
	while ((de=readdir(d))!=NULL)
	{
		char *sid=NULL;
		char *sfrom=NULL;
		char *sto=NULL;
		char *sstatus=NULL;
		time_t smodzeit=-1;
		time_t sorigzeit=-1;

		char buf[1024];
		FILE *fin;
		pid_t pid;

		if (strncmp(de->d_name,"st",2))
			continue; /* falsches Paket */

		pid=strtol(de->d_name+2,NULL,10);
		if (!pid)
			continue;
		errno=0;
		if (kill(pid,0) == 0  || errno==EPERM)
			/* der Prozess existiert noch und schreibt vielleicht 
			 * gerade noch etwas in die Datei */
			continue;
		/* ok, this is our file */
		sprintf(fname,"%s%lu/%s",box_conf.spool_dir,(ULONG) unr, de->d_name);

		fin=fopen(fname,"rb");
		if (!fin)
			continue;
		unlink(fname); /* Racecondition vermeiden */
		while (fgets(buf,sizeof(buf),fin))
		{
			char *p=strpbrk(buf,"\r\n");
			if (p) *p=0;
			switch(*buf)
			{
			case '#':
				if (sid)
				{
					if (smodzeit!=(time_t) -1 && sfrom && sstatus && sto)
						ok&=r_st_write_one_status_to_user(f,sid,sfrom,sto,
							sstatus,smodzeit,sorigzeit);
					free(sfrom);
					free(sto);
					free(sstatus);
					free(sid);
					smodzeit=-1;
					sorigzeit=-1;
				}
				sid=XSTRDUP(buf+1);
				break;
			case 'V':
				sfrom=XSTRDUP(buf+1);
				break;
			case 'A':
				sto=XSTRDUP(buf+1);
				break;
			case 'B':
				sstatus=XSTRDUP(buf+1);
				break;
			case 'E':
				sorigzeit=strtol(buf+1,NULL,10);
				break;
			case 'M':
				smodzeit=strtol(buf+1,NULL,10);
				break;
			}
		}
		fclose(fin);
		if (sid)
		{
			if (smodzeit!=(time_t) -1 && sfrom && sstatus && sto)
				ok&=r_st_write_one_status_to_user(f,sid,sfrom,sto,
					sstatus,smodzeit,sorigzeit);
			free(sfrom);
			free(sto);
			free(sstatus);
			free(sid);
		}
	}
	closedir(d);
	return ok;


}

static BOOLEAN
r_st_write_one_status_to_user(FILE *f, const char *id, const char *from,
	const char *to, const char *status, time_t modzeit, time_t origzeit)
{
	/* 
	 * Achtung: Die letzte Zeile eines Datensatzes mu immer die C-Zeile
	 * sein. Die Reihenfolge mu # "AV" E "BMC" sein.
	 */
	char buf[1024];
	BOOLEAN found=FALSE;
	fseek(f,0,SEEK_SET);
	while (fgets(buf,sizeof(buf),f))
	{
		if (*buf=='#')
		{
			char *p;
			p=strchr(buf,'\n');
			if (p)
			{
				*p--=0;
				if (p >= buf && *p=='\r')
					*p=0;
			}
			if (!stricmp(buf+1,id))
				found=TRUE;
		}
		else if (*buf=='E')
		{
			if (found)
			{
				if (fseek(f,0,SEEK_CUR))
					return FALSE;
				putc('B',f);
				putc(*status,f);
				fputs("\r\nM",f);
				fprintf(f,"%10.10lu",(ULONG) modzeit);
				fputs("\r\nC",f);
				fprintf(f,"%10.10lu\r\n",(ULONG) time(NULL));
				return TRUE;
			}
		}
	}
	if (fseek(f,0,SEEK_END))
		return FALSE;
	fprintf(f,
		"#%s\r\n"
		"V%s\r\n"
		"A%s\r\n"
		"E%10.10lu\r\n"
		"B%c\r\n"
		"M%10.10lu\r\n"
		"C%10.10lu\r\n",
			id, 
			from ? from : "unbekannt", 
			to, 
			(ULONG) origzeit,
			*status,
			(ULONG) modzeit,
			(ULONG) time(NULL));
	return TRUE;
}

BOOLEAN
r_st_export(FILE *out, time_t seit)
{
	char *id=NULL;
	char status=0;
	time_t orig=-1;
	time_t setz=-1;
	time_t mod=-1;
	char fname[PATH_MAX];
	char buf[1024];
	char ozeit[20];
	char mzeit[20];
	FILE *f;

	/* Alle fr diesen User noch bereitliegenden Statuspakete importieren */
	r_st_write_status_to_user(0,NULL,NULL,NULL,NULL,0,0);

	sprintf(fname,"%s%lu.status",box_conf.tausch_dir,(ULONG) user.nummer);
	f=fopen(fname,"rb");
	if (!f)
		return TRUE; /* kein Status da */

	if (0!=fd_read_lock(fileno(f), 0,SEEK_SET,0))
	{
		/* irgendjemand schreibt da gerade */
		fclose(f);
		return FALSE;
	}

	/* 
	 * Der Code hngt davon ab, da "C" immer die letzte Zeile ist! 
	 */
	while (fgets(buf,sizeof(buf),f))
	{
		char *p=strpbrk(buf,"\r\n");
		if (p) *p=0;
		switch(*buf)
		{
		case '#': 
			if (id)
				free(id);
			id=XSTRDUP(buf+1); break;
		case 'B': status=buf[1]; break;
		case 'E': orig=strtol(buf+1,NULL,10); break;
		case 'M': mod=strtol(buf+1,NULL,10); break;
		case 'C': setz=strtol(buf+1,NULL,10);
			if (id && status && setz!=-1 && setz > seit && mod!=-1 && orig!=-1)
			{
				time_t_to_mtzeit(&mod,mzeit);
				time_t_to_mtzeit(&orig,ozeit);
				fprintf(out,
					"#%s\r\n"
					"B%c%s\r\n"
					"E%s\r\n",
						id,status,mzeit,ozeit);
			}
			break;
		}
	}
	fclose(f);
	if (id)
		free(id);
	return TRUE;
}

static BOOLEAN
r_st_expire_user(USERNR unr)
{
	char fname[PATH_MAX];
	char restbuf[16384];
	char *wp;
	char buf[1024];
	FILE *alt;
	FILE *neu;
	BOOLEAN too_long=FALSE;
	size_t bytes;
	time_t vglzeit=time(NULL);
	BOOLEAN deleted_at_least_one=FALSE;

	sprintf(fname,"%s%lu.status",box_conf.tausch_dir,(ULONG) unr);
	alt=fopen(fname,"r+b");
	if (!alt)
		return TRUE; /* wo nichts zu lschen ist ... */

	if (0!=fd_write_lock(fileno(alt), 0,SEEK_SET,0))
	{
		fclose(alt);
		return FALSE;
	}
	sprintf(fname,"%s%lu.nstatus",box_conf.tausch_dir,(ULONG) unr);
	neu=xfopen(fname,"w+b");
	unlink(fname);

	wp=restbuf;
	while (fgets(buf,sizeof(buf),alt))
	{
		size_t l;
		char *p=strpbrk(buf,"\r\n");
		if (p) *p=0;
		l=strlen(buf);
		if (wp-restbuf+l+2<sizeof(restbuf))
		{
			strcpy(wp,buf);
			wp+=l;
			strcpy(wp,"\r\n");
			wp+=2;
		}
		else
			too_long=TRUE;
		if (*buf=='C')
		{
			time_t gesetzt=strtol(buf+1,NULL,10);
			if (!too_long && (vglzeit - gesetzt)/86400L < 90)
			{
				if (fwrite(restbuf,wp-restbuf,1,neu)!=1)
				{
					/* grrr ... */
					fclose(neu);
					fclose(alt);
					return FALSE; /* vielleicht ist ja gleich mehr Platz da */
				}
			}
			else
				deleted_at_least_one=TRUE;
			wp=restbuf;
			too_long=FALSE;
		}
	}
	if (!deleted_at_least_one)
	{
		fclose(neu);
		fclose(alt);
		return TRUE;
	}
	/* nun von neu nach alt kopieren. Warum nicht rename? Mgliche 
	 * Racecondition - da existiert einen Moment die Datei nicht 
	 * und da knnte jemand sie neu anlegen.
	 */
	fseek(neu,0,SEEK_SET);
	fseek(alt,0,SEEK_SET);
	while ((bytes=fread(restbuf,1,sizeof(restbuf),neu))!=0)
	{
		fwrite(restbuf,1,bytes,alt);
	}
	fflush(alt);
	ftruncate(fileno(alt),ftell(neu));
	fclose(neu);
	fclose(alt);
	return TRUE;
}

static BOOLEAN
r_st_expire_global(void)
{
	char fname[PATH_MAX];
	FILE *neu;
	char buf[1024];
	char id[1024];
	char restbuf[16384];
	char *wp=restbuf;
	BOOLEAN too_long=FALSE;
	BOOLEAN deleted_at_least_one=FALSE;
	time_t vglzeit=time(NULL);
	size_t bytes;

	if (!r_st_open_global(TRUE))
		return FALSE;
	sprintf(fname,"%smailstatus.neu",box_conf.var_dir);
	neu=xfopen(fname,"w+b");
	unlink(fname);
	while (fgets(buf,sizeof(buf),s_file))
	{
		char *p=strpbrk(buf,"\r\n");
		size_t l;
		if (p) *p=0;
		l=strlen(buf);

		if (wp-restbuf+l+2<sizeof(restbuf))
		{
			memcpy(wp,buf,l);
			wp+=l;
			memcpy(wp,"\r\n",2);
			wp+=2;
		}
		else
			too_long=TRUE;
		switch(*buf)
		{
		case '#': 
			strcpy(id,buf+1); 
			break;
		case 'e': 
			{
				time_t angelegt=strtol(buf+1,NULL,10);
				/* ### konfigurierbar machen? */
				if ((vglzeit-angelegt)/86400L > 90 || too_long)
					deleted_at_least_one=TRUE;
				else
				{

					/* die behalten wir also */
					if (1!=fwrite(restbuf,wp-restbuf,1,neu))
					{
						/* Oh weh */
						syslog(LOG_WARNING,
							"cannot write global mailstatus database");
						fclose(neu);

						/* Und Neuaufbau */
						fclose(neu);
						r_st_close_global(NULL);

						return TRUE; /* es mu weitergehen! */
					}
				}
			}
			wp=restbuf;
			too_long=FALSE;
			break;
		}
	}
	if (!deleted_at_least_one 
		&& !fileexist("tom/var/mailstatus.force")
		&& !fileexist("tom/system/mailstatus.force"))
	{
		fclose(neu);
		r_st_close_global(NULL);
		return TRUE;
	}
	unlink("tom/var/mailstatus.force");
	unlink("tom/system/mailstatus.force");
	/* nun von neu nach alt kopieren. Warum nicht rename? Mgliche 
	 * Racecondition - da existiert einen Moment die Datei nicht 
	 * und da knnte jemand sie neu anlegen.
	 */
	fseek(neu,0,SEEK_SET);
	fseek(s_file,0,SEEK_SET);
	while ((bytes=fread(restbuf,1,sizeof(restbuf),neu))!=0)
	{
		fwrite(restbuf,1,bytes,s_file);
	}
	fclose(neu);

	fflush(s_file);
	ftruncate(fileno(s_file),ftell(s_file));
	fseek(s_file,0,SEEK_SET);
	r_st_rebuild_db(s_file);

	r_st_close_global(NULL);
	return TRUE;
}

BOOLEAN
r_st_expire(void)
{
	struct expire
	{
		ULONG nr;
		struct expire *next;
	} *liste;
	LISTUSER *lu;
	int sleeptime=2;

	/* Alles Liegengebliebene erst mal importieren */
	r_st_newmail(NULL,NULL,NULL,0);
	r_st_set_status(NULL,NULL,NULL,NULL,0);

	/* 
	 * Beide Datenbanken sind unabhngig, so da wir nicht zwei 
	 * gleichzeitig locken mssen.
	 * Zuerst mal eine Liste aller zu lockenden Datenbanken aufbauen.
	 */

	liste=XMALLOC(sizeof(struct expire));
	liste->nr=0; /* Flag fr Globaldatenbank */
	liste->next=NULL;

	lu=ub_get_first();
	while (lu)
	{
		struct expire *n=XMALLOC(sizeof(struct expire));
		n->nr=lu->nummer;
		n->next=liste;
		liste=n;
		lu=lu->next;
	}

	while (liste)
	{
		struct expire *akt;
		struct expire *last=NULL;
		akt=liste;
		while (akt)
		{
			BOOLEAN done;
			if (akt->nr==0)
				done=r_st_expire_global();
			else
				done=r_st_expire_user(akt->nr);
			if (done)
			{
				if (last)
				{
					last->next=akt->next;
					free(akt);
					akt=last->next;
				}
				else
				{
					liste=akt->next;
					free(akt);
					akt=liste;
				}
			}
			else
			{
				last=akt;
				akt=akt->next;
			}
		}
		/* be friendly to the poor machine */
		if (liste)
		{
			sleep(sleeptime);
			/* das sind 30 Minuten, und mte fr jeden Poll ausreichen */
			if (sleeptime<30*60)
				sleeptime*=2;
		}
	}
	return TRUE;
}


static void
r_st_rebuild_db(FILE *in)
{
	char db_fname[PATH_MAX];
	char db_fname2[PATH_MAX];
	char buf[1024];
	off_t pos;
	ULONG crc32;

	sprintf(db_fname,"%smailstatus.dir",box_conf.var_dir);
	sprintf(db_fname2,"%smailstatus.pag",box_conf.var_dir);
	unlink(db_fname);
	unlink(db_fname2);

	s_dbm = dbm_prep(db_fname,db_fname2,O_RDWR|O_CREAT,0640);
	XNULLCHECK(s_dbm); /* This must not happen */

	pos=0;
	while (fgets(buf,sizeof(buf),in))
	{
		if (*buf=='#')
		{
			char *p=buf+strlen(buf)-1;
			datum key,set;
			if (*p=='\n')
				*p--=0;
			if (*p=='\r')
				*p--=0;
			p = normalize_id (buf+1);

			crc32=crc32ansi(p,strlen(p));

			key.dptr = (char *) &crc32;
			key.dsize = sizeof (crc32);

			set.dptr=(char *) &pos;
			set.dsize=sizeof(pos);
			if (dbm_store (s_dbm, key, set, DBM_INSERT)==-1)
			{
				syslog(LOG_WARNING,
					"cannot write index to global mailstatus database");
				/* Hm? */
				free(p);
				return;
			}
			free(p);

		}
		pos=ftell(in);
	}
}
